import random
from ..action import *
from ..structure import *

async def get_action(M, max_time_for_dec, epochs, X_train, Y_train, simulation_score):
    all_action_seq = Action.generate_all_actions(M)
    if len(all_action_seq) == 0:
        print("Error no actions avaible")
    return random.choice(all_action_seq), 0, 0


import asyncio
import sys
sys.path.append('.')
sys.path.append('../')
import growingnn as gnn
import unittest
import pandas as pd
import random
from unittest.mock import MagicMock

class TestHistory(unittest.TestCase):

    def setUp(self):
        self.keys = ['loss', 'accuracy']
        self.history = gnn.History(self.keys)

    def test_init(self):
        self.assertDictEqual(self.history.Y, {'loss': [], 'accuracy': []})
        self.assertEqual(self.history.last_img_id, 0)
        self.assertEqual(self.history.description, "\ndescription_of_training_process: \n")
        self.assertEqual(self.history.best_train_acc, 0.0)
        self.assertEqual(self.history.best_test_acc, 0.0)

    def test_get_length(self):
        self.assertEqual(self.history.get_length(), 0)
        self.history.Y['loss'] = [0.3, 0.4, 0.5]
        self.assertEqual(self.history.get_length(), 3)

    def test_append(self):
        self.history.append('loss', 0.3)
        self.history.append('accuracy', 0.9)
        self.assertEqual(self.history.Y['loss'], [0.3])
        self.assertEqual(self.history.Y['accuracy'], [0.9])

    def test_merge(self):
        new_hist = gnn.History(self.keys)
        new_hist.append('loss', 0.4)
        new_hist.append('accuracy', 0.8)
        self.history.Y['loss'] = [0.3]
        self.history.Y['accuracy'] = [0.9]
        self.history.merge(new_hist)
        self.assertEqual(self.history.Y['loss'], [0.3, 0.4])
        self.assertEqual(self.history.Y['accuracy'], [0.9, 0.8])

    def test_learning_capable_below_threshold(self):
        self.history.Y['loss'] = [0.1] * 9
        self.assertTrue(self.history.learning_capable())
        
    def test_get_last(self):
        self.history.Y['loss'] = [0.3, 0.4, 0.5]
        self.assertEqual(self.history.get_last('loss'), 0.5)


if __name__ == '__main__':
    unittest.main()



import time
import random
import math
from ..action import *
from ..structure import *

UCB1_CONTS = 2
DEEPTH = 2

def protected_divide(a,b):
    LARGE_MAX = 2**128
    if a>LARGE_MAX or b > LARGE_MAX:
        return a//b 
    return a/b

class TreeNode:
    def __init__(self, _parent, _action, _M, _epochs, _X_train, _Y_train, _simulation_score):
        self.parent = _parent
        self.action = _action
        self.X_train = _X_train
        self.Y_train = _Y_train
        self.epochs = _epochs
        self.action = _action
        self.simulation_score = _simulation_score
        self.M = _M
        self.childNodes = []
        self.value = 0
        self.visit_counter = 0

    def expand(self):
#        all_action_seq = self.M.generate_all_possible_new_layers()
        all_action_seq = Action.generate_all_actions(self.M)
        for action in all_action_seq:
            M_copy = self.M.deepcopy()
            action.execute(M_copy)
            #M_copy.add_layer(action[0], action[1])
            new_node = TreeNode(self, action, M_copy, self.epochs, self.X_train, self.Y_train, self.simulation_score)
            self.childNodes.append(new_node)

    def scoreFun(M, epochs, X_train, Y_train, simulation_score):
        #print("X_train: ", X_train.shape)
        #print("Y_train: ", Y_train.shape)
        acc, history = M.gradient_descent(X_train, Y_train, epochs, LearningRateScheduler(LearningRateScheduler.CONSTANT, 0.1) , True)
        #print("loss in MCTS: ", history.get_last('loss'), " off: ", max_loss - history.get_last('loss'))
        #return max(1.e-17, max_loss - history.get_last('loss'))
        return simulation_score.grade(acc, history)
    
    def rollout(self): 
        M_copy = self.M.deepcopy()
        #all_action_seq = M_copy.generate_all_possible_new_layers()
        
        #Action.generate_all_actions(self.M)
        deepth = DEEPTH
        while deepth > 0:
            all_action_seq = Action.generate_all_actions(M_copy)
            if len(all_action_seq) == 0: break
            choosen_action = random.choice(all_action_seq)
            choosen_action.execute(M_copy)
            #M_copy.add_layer(action_seq[0], action_seq[1])
            all_action_seq.remove(choosen_action)
            new_action_seq = []
            for check_action in all_action_seq:
                if check_action.can_be_infulenced(choosen_action) == False:
                    new_action_seq.append(check_action)
            all_action_seq = new_action_seq
            deepth -= 1

        score = TreeNode.scoreFun(M_copy, self.epochs, self.X_train, self.Y_train, self.simulation_score)
        return score

    def get_best_child(self):
        USB1 = lambda node : node.value + UCB1_CONTS*protected_divide(math.log(node.parent.visit_counter),node.visit_counter)
        gradeChild = lambda node : float('inf') if node.visit_counter==0 else USB1(node)
        bestScore = 0
        bestChild = None
        for child in self.childNodes:
            score = gradeChild(child)
            if score > bestScore:
                bestScore = score
                bestChild = child
        return bestChild
        
    def is_leaf(self): 
        return len(self.childNodes)==0

    def get_depth(self):
        res = 0 
        for child in self.childNodes:
            res = max(res,child.get_depth())
        return res + 1

    def kill(self):
        for child in self.childNodes:
            child.kill()
        del self

    def __str__(self):
        result = "node: value:"+ str(self.value) + " visit_counter: "+str(self.visit_counter)+ "\n"
        for child in self.childNodes:
            result += child.__str__()+ "\n"
        return result
    
async def get_action(M, max_time_for_dec, epochs, X_train, Y_train, simulation_score):
    size_of_changes = len(Action.generate_all_actions(M))
    #size_of_changes = len(M.generate_all_possible_new_layers())
    if size_of_changes == 0: 
        print("Error")
        return None,0
    root = TreeNode(None, None, M, epochs, X_train, Y_train, simulation_score)
    deadline = time.time() + max_time_for_dec
    deepth = 0
    rollouts = 0
    while time.time() < deadline or rollouts <= size_of_changes:
        _, deepth, r = simulate(root)
        rollouts += r
    if time.time() > deadline: 
        print("More time was needed to analise all possiblites at least once")
    best_action = root.get_best_child().action
    root.kill()
    return best_action, deepth, rollouts

def simulate(node, deepth = 0, rollouts = 0):
    if node.is_leaf():
        if node.visit_counter==0:
            new_value = node.rollout()
            #print("deepth: ", deepth, " rollout score: ", new_value, )
            node.value = new_value
            node.visit_counter += 1
            rollouts += 1
            return node.value, deepth, rollouts
        else:
            node.expand()
    child = node.get_best_child()
    if child == None: return node.value, deepth, rollouts
    new_value, deepth, rollouts = simulate(child, deepth + 1, rollouts)        
    node.value += new_value
    node.visit_counter += 1
    return node.value, deepth, rollouts





error_clip_range = 600 # In back propagation we are clipping the error range to prevent exploding gradients
weights_clip_range = 400 # After weight update we are clipping all weights to prevent exploding gradients


import asyncio
import sys
sys.path.append('.')
sys.path.append('../')
import growingnn as gnn
import unittest
import pandas as pd
import random
from unittest.mock import MagicMock

class TestLayer(unittest.TestCase):

    def setUp(self):
        self.layer = gnn.Layer(1, None, 10, 5, None)
        self.layer.act_fun = gnn.structure.Activations.ReLu
        self.layer.is_ending = True


    def test_init(self):
        self.assertEqual(self.layer.id, 1)
        self.assertIsNone(self.layer.model)
        self.assertEqual(self.layer.input_size, 10)
        self.assertEqual(self.layer.neurons, 5)
        self.assertEqual(self.layer.input_layers_ids, [])
        self.assertEqual(self.layer.output_layers_ids, [])
        self.assertEqual(self.layer.f_input, [])
        self.assertEqual(self.layer.b_input, [])
        self.assertEqual(self.layer.reshspers, {})

    def test_connect_input(self):
        self.layer.model = MagicMock()
        self.layer.model.get_layer.return_value = MagicMock(output_layers_ids=[])
        self.layer.connect_input(2)
        self.assertIn(2, self.layer.input_layers_ids)
        self.layer.model.get_layer.assert_called_with(2)
        self.layer.model.get_layer.return_value.connect_output.assert_called_with(1)

    def test_connect_output(self):
        self.layer.model = MagicMock()
        self.layer.model.get_layer.return_value = MagicMock(input_layers_ids=[])
        self.layer.connect_output(2)
        self.assertIn(2, self.layer.output_layers_ids)
        self.layer.model.get_layer.assert_called_with(2)
        self.layer.model.get_layer.return_value.connect_input.assert_called_with(1)

    def test_disconnect(self):
        self.layer.input_layers_ids = [2, 3]
        self.layer.output_layers_ids = [4, 5]
        self.layer.disconnect(3)
        self.assertNotIn(3, self.layer.input_layers_ids)
        self.assertEqual(self.layer.output_layers_ids, [4, 5])
        self.layer.disconnect(5)
        self.assertEqual(self.layer.output_layers_ids, [4])


    def test_forward_prop(self):
        self.layer.input_layers_ids = [2, 3]
        self.layer.model = MagicMock()
        result1 = self.layer.forward_prop(gnn.np.ones((10, 1)))
        result2 = self.layer.forward_prop(gnn.np.ones((10, 1)))
        self.assertIsNone(result1)
        self.assertIsNotNone(result2)
        self.assertEqual(len(self.layer.f_input), 0)

    def test_back_prop(self):
        self.layer.output_layers_ids = [2, 3]
        self.layer.model = MagicMock()

        self.layer.E = gnn.np.ones((5, 1))
        self.layer.b_input = [gnn.np.ones((5, 1)), gnn.np.ones((5, 1))]

        self.layer.forward_prop(gnn.np.ones((10, 1)))
        self.layer.back_prop(gnn.np.ones((5, 1)), 1, 0.01)

        self.assertEqual(len(self.layer.b_input), 0)
        self.assertTrue(hasattr(self.layer, 'dW'))
        self.assertTrue(hasattr(self.layer, 'dB'))

if __name__ == '__main__':
    unittest.main()



import asyncio
import sys
sys.path.append('.')
sys.path.append('../')
import growingnn as gnn
import unittest
import pandas as pd
import random
from unittest.mock import MagicMock

class TestSimulationScore(unittest.TestCase):

    def setUp(self):
        self.global_history_mock = MagicMock()
        self.history_mock = MagicMock()

    def test_new_max_loss(self):
        self.global_history_mock.Y = {'loss': [0.3, 0.5, 0.2]}
        sim_score = gnn.Simulation_score(gnn.Simulation_score.LOSS)
        sim_score.new_max_loss(self.global_history_mock)
        self.assertEqual(sim_score.max_loss, 0.5)

    def test_grade_accuracy_mode(self):
        sim_score = gnn.Simulation_score(gnn.Simulation_score.ACCURACY)
        result = sim_score.grade(0.9, self.history_mock)
        self.assertEqual(result, 0.9)

    def test_grade_loss_mode(self):
        sim_score = gnn.Simulation_score(gnn.Simulation_score.LOSS)
        sim_score.max_loss = 0.5
        self.history_mock.get_last.return_value = 0.3
        result = sim_score.grade(0.9, self.history_mock)
        self.assertEqual(result, 0.2)

    def test_grade_loss_mode_min_threshold(self):
        sim_score = gnn.Simulation_score(gnn.Simulation_score.LOSS)
        sim_score.max_loss = 0.1
        self.history_mock.get_last.return_value = 0.2
        result = sim_score.grade(0.9, self.history_mock)
        self.assertEqual(result, 1.e-17)

    def test_grade_accuracy(self):
        score = gnn.Simulation_score(mode=gnn.Simulation_score.ACCURACY)
        result = score.grade(acc=0.9, history=None)
        self.assertEqual(result, 0.9)
        
    def test_grade_loss(self):
        score = gnn.Simulation_score(mode=gnn.Simulation_score.LOSS)
        global_history = MagicMock()
        global_history.Y = {'loss': [0.5, 0.4, 0.3]}
        score.new_max_loss(global_history)
        history = MagicMock()
        history.get_last.return_value = 0.3
        result = score.grade(acc=None, history=history)
        expected = 0.5 - 0.3
        self.assertAlmostEqual(result, expected)

if __name__ == '__main__':
    unittest.main()



import unittest

# Define the test file names
test_files = [
    "testHistory.py",
    "testLayer.py",
    "testLRscheduler.py",
    "testModel.py",
    "testSimulationScheduler.py",
    "testSimulationScore.py",
    "testStorage.py",
    "unittests_autogenerated.py",
    "unittests_basic.py"
]

# Create a test loader
loader = unittest.TestLoader()

# Create an empty test suite
suite = unittest.TestSuite()

# Add tests from each test file to the test suite
for test_file in test_files:
    # Discover tests in the test file
    discovered_tests = loader.discover('./tests/', pattern=test_file)
    # Add discovered tests to the test suite
    suite.addTests(discovered_tests)

# Create a test runner that will output the results to the console
runner = unittest.TextTestRunner()

# Run the test suite
if __name__ == '__main__':
    runner.run(suite)



from ..helpers import *
from ..action import *
from ..structure import *
from sklearn.decomposition import PCA

# Simulation set

def create_simulation_set_SAMLE(x,y, amount = 20):
    #set_size_ratio = float(amount) / float(max(np.shape(y)))
    x_sim, y_sim = protected_sampling(x, y, amount)
    #_, x_sim, _, y_sim = train_test_split(x, y, test_size=set_size_ratio)
    return x_sim, y_sim

def create_simulation_set_PCA(x,y, amount = 10):
    #print("1 x shape: ", x.shape)
    #print("22: x: ", x.shape[0], " y: ", y.shape[0])
    reverse_axes = False
    if x.shape[0] != y.shape[0]:
        x = np.swapaxes(x, 0, 1)
        reverse_axes = True
        #print("2 x shape: ", x.shape)
    if y.shape[0] < 100: return create_simulation_set_SAMLE(x,y,amount)
    if x.shape[0] != y.shape[0]: print("error wrong dimensions !")
    pca = PCA(amount)
    classes = {}
    x_result = []
    y_result = []
    for i in range(x.shape[0]):
        if not y[i] in classes.keys():
            classes[y[i]] = []
        classes[y[i]].append(x[i])
    for c in classes:
        x_per_class = np.array(classes[c])
        needed_shape = (amount,) + x_per_class.shape[1:]
        n = np.prod(x_per_class.shape[1:]) # obliczamy długość wektora spłaszczonego
        x_per_class_flat = x_per_class.reshape(x_per_class.shape[0], n)
        x_per_class_flat = x_per_class_flat.T
        sim_x_per_class_flat = pca.fit_transform(x_per_class_flat)
        sim_x_per_class = sim_x_per_class_flat.reshape(needed_shape)
        x_result.append(sim_x_per_class)
        y_result.append(np.ones((amount,)) * int(c))
    x_result = np.array(x_result)
    y_result = np.array(y_result).astype(int)
    x_result = np.concatenate(x_result, axis=0)
    y_result = np.concatenate(y_result, axis=0)
    print("3 x shape: ", x.shape)
    if reverse_axes:
        x = np.swapaxes(x, 0, 1)
        print("4 x shape: ", x.shape)
        print("x_result: ", x_result.shape)
        print("y_result: ", y_result.shape)
        x_result = np.swapaxes(x_result, 0, 1)
    print("x_result: ", x_result.shape)
    print("y_result: ", y_result.shape)
    return x_result, y_result